@@ -33,7 +33,9 @@
 #include "editor-session-private.h"
 #include "editor-window.h"
 
-#define METATDATA_CURSOR    "metadata::gnome-text-editor-cursor"
+#define METADATA_CURSOR     "metadata::gte-cursor"
+#define METADATA_SPELLING   "metadata::gte-spelling"
+#define METADATA_SYNTAX     "metadata::gte-syntax"
 #define TITLE_LAST_WORD_POS 20
 #define TITLE_MAX_LEN       100
 
@@ -44,8 +46,8 @@ struct _EditorDocument
   EditorBufferMonitor          *monitor;
   GtkSourceFile                *file;
   gchar                        *draft_id;
-  GtkTextTag                   *line_spacing_tag;
   const GtkSourceEncoding      *encoding;
+  GError                       *last_error;
 
   EditorSpellChecker           *spell_checker;
   EditorTextBufferSpellAdapter *spell_adapter;
@@ -59,6 +61,8 @@ struct _EditorDocument
   guint                         needs_autosave : 1;
   guint                         was_restored : 1;
   guint                         externally_modified : 1;
+  guint                         suggest_admin : 1;
+  guint                         load_failed : 1;
 };
 
 typedef struct
@@ -77,6 +81,7 @@ typedef struct
   gint64           draft_modified_at;
   gint64           modified_at;
   guint            n_active;
+
   guint            highlight_syntax : 1;
   guint            has_draft : 1;
   guint            has_file : 1;
@@ -90,13 +95,158 @@ enum {
   PROP_BUSY_PROGRESS,
   PROP_EXTERNALLY_MODIFIED,
   PROP_FILE,
+  PROP_HAD_ERROR,
   PROP_SPELL_CHECKER,
+  PROP_SUGGEST_ADMIN,
   PROP_TITLE,
   N_PROPS
 };
 
+enum {
+  SAVE,
+  N_SIGNALS
+};
+
 static GParamSpec *properties [N_PROPS];
 static GSettings *shared_settings;
+static guint signals [N_SIGNALS];
+
+static gboolean
+is_admin (EditorDocument *self)
+{
+  g_autofree char *uri = NULL;
+  GFile *file;
+
+  g_assert (EDITOR_IS_DOCUMENT (self));
+
+  if (!(file = editor_document_get_file (self)))
+    return FALSE;
+
+  if (g_file_is_native (file))
+    return FALSE;
+
+  uri = g_file_get_uri (file);
+
+  return g_str_has_prefix (uri, "admin:///");
+}
+
+static void
+editor_document_track_error (EditorDocument *self,
+                             const GError   *error)
+{
+  g_assert (EDITOR_IS_DOCUMENT (self));
+
+  g_clear_error (&self->last_error);
+  self->last_error = error ? g_error_copy (error) : NULL;
+
+  self->load_failed = error != NULL;
+
+  editor_buffer_monitor_set_failed (self->monitor, self->load_failed);
+
+  if (error == NULL)
+    {
+      self->suggest_admin = FALSE;
+      g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_SUGGEST_ADMIN]);
+      g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_HAD_ERROR]);
+      return;
+    }
+
+  if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED))
+    {
+      if (!is_admin (self))
+        {
+          self->suggest_admin = TRUE;
+          g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_SUGGEST_ADMIN]);
+        }
+    }
+
+  g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_HAD_ERROR]);
+}
+
+static void
+check_error (GObject      *object,
+             GAsyncResult *result,
+             gpointer      user_data)
+{
+  GFile *file = (GFile *)object;
+  g_autoptr(GError) error = NULL;
+
+  if (!g_file_set_attributes_finish (file, result, NULL, &error))
+    g_warning ("%s", error->message);
+}
+
+void
+_editor_document_persist_syntax_language (EditorDocument *self,
+                                          const char     *language_id)
+{
+  g_autoptr(GFileInfo) info = NULL;
+  GFile *file;
+
+  g_assert (EDITOR_IS_DOCUMENT (self));
+
+  /* Only persist the metadata if we have a backing file */
+  if (!(file = editor_document_get_file (self)) || !g_file_is_native (file))
+    return;
+
+  info = g_file_info_new ();
+  g_file_info_set_attribute_string (info, METADATA_SYNTAX, language_id ? language_id : "");
+  g_file_set_attributes_async (file, info, G_FILE_QUERY_INFO_NONE,
+                               G_PRIORITY_DEFAULT, NULL, check_error, NULL);
+}
+
+static void
+on_spelling_language_changed_cb (EditorDocument     *self,
+                                 GParamSpec         *pspec,
+                                 EditorSpellChecker *spell_checker)
+{
+  g_autoptr(GFileInfo) info = NULL;
+  const char *language_id;
+  GFile *file;
+
+  g_assert (EDITOR_IS_DOCUMENT (self));
+  g_assert (EDITOR_IS_SPELL_CHECKER (spell_checker));
+
+  /* Only persist the metadata if we have a backing file */
+  if (!(file = editor_document_get_file (self)) || !g_file_is_native (file))
+    return;
+
+  /* Ignore if there is nothing to set */
+  if (!(language_id = editor_spell_checker_get_language (spell_checker)))
+    return;
+
+  info = g_file_info_new ();
+  g_file_info_set_attribute_string (info, METADATA_SPELLING, language_id);
+  g_file_set_attributes_async (file, info, G_FILE_QUERY_INFO_NONE,
+                               G_PRIORITY_DEFAULT, NULL, check_error, NULL);
+}
+
+static void
+editor_document_emit_save (EditorDocument *self)
+{
+  g_assert (EDITOR_IS_DOCUMENT (self));
+
+  g_signal_emit (self, signals [SAVE], 0);
+}
+
+static GtkSourceLanguage *
+guess_language (GtkSourceLanguageManager *manager,
+                const char               *filename,
+                const char               *content_type)
+{
+  /* Apply content-type overrides */
+  if (filename != NULL && content_type != NULL)
+    {
+      /* shared-mime-info calls it text/x-python rather than text/x-python3
+       * if the shebang is something like "env python" vs "env python3". We
+       * always want to default to Python 3.
+       */
+      if (g_str_has_suffix (filename, ".py") &&
+          g_str_equal (content_type, "text/x-python"))
+        content_type = "text/x-python3";
+    }
+
+  return gtk_source_language_manager_guess_language (manager, filename, content_type);
+}
 
 static void
 load_free (Load *load)
@@ -115,17 +265,6 @@ save_free (Save *save)
   g_slice_free (Save, save);
 }
 
-static void
-editor_document_apply_spacing (EditorDocument *self)
-{
-  GtkTextIter begin, end;
-
-  g_assert (EDITOR_IS_DOCUMENT (self));
-
-  gtk_text_buffer_get_bounds (GTK_TEXT_BUFFER (self), &begin, &end);
-  gtk_text_buffer_apply_tag (GTK_TEXT_BUFFER (self), self->line_spacing_tag, &begin, &end);
-}
-
 static GMountOperation *
 editor_document_mount_operation_factory (GtkSourceFile *file,
                                          gpointer       user_data)
@@ -148,6 +287,11 @@ editor_document_load_notify_completed_cb (EditorDocument *self,
   g_assert (EDITOR_IS_DOCUMENT (self));
   g_assert (G_IS_TASK (task));
 
+  self->loading = FALSE;
+
+  if (!g_task_had_error (task))
+    editor_document_track_error (self, NULL);
+
   session = editor_application_get_session (EDITOR_APPLICATION_DEFAULT);
   _editor_session_document_seen (session, self);
 }
@@ -216,7 +360,7 @@ editor_document_guess_content_type (EditorDocument *self)
 
   content_type = g_content_type_guess (filename, (const guchar *)content, strlen (content), &uncertain);
   manager = gtk_source_language_manager_get_default ();
-  language = gtk_source_language_manager_guess_language (manager, filename, content_type);
+  language = guess_language (manager, filename, content_type);
 
   if (language)
     gtk_source_buffer_set_language (GTK_SOURCE_BUFFER (self), language);
@@ -401,12 +545,6 @@ editor_document_constructed (GObject *object)
                                 self->spell_adapter, "enabled",
                                 G_SETTINGS_BIND_GET,
                                 apply_spellcheck_mapping, NULL, self, NULL);
-
-  self->line_spacing_tag = gtk_text_buffer_create_tag (GTK_TEXT_BUFFER (self),
-                                                       NULL,
-                                                       "pixels-below-lines", 2,
-                                                       NULL);
-  editor_document_apply_spacing (self);
 }
 
 static void
@@ -418,6 +556,7 @@ editor_document_finalize (GObject *object)
   g_clear_object (&self->file);
   g_clear_object (&self->spell_checker);
   g_clear_object (&self->spell_adapter);
+  g_clear_error (&self->last_error);
   g_clear_pointer (&self->draft_id, g_free);
 
   G_OBJECT_CLASS (editor_document_parent_class)->finalize (object);
@@ -433,6 +572,14 @@ editor_document_get_property (GObject    *object,
 
   switch (prop_id)
     {
+    case PROP_SUGGEST_ADMIN:
+      g_value_set_boolean (value, self->suggest_admin);
+      break;
+
+    case PROP_HAD_ERROR:
+      g_value_set_boolean (value, _editor_document_had_error (self));
+      break;
+
     case PROP_BUSY:
       g_value_set_boolean (value, editor_document_get_busy (self));
       break;
@@ -500,6 +647,20 @@ editor_document_class_init (EditorDocumentClass *klass)
   buffer_class->insert_text = editor_document_insert_text;
   buffer_class->delete_range = editor_document_delete_range;
 
+  properties [PROP_SUGGEST_ADMIN] =
+    g_param_spec_boolean ("suggest-admin",
+                          "Suggest Admin",
+                          "Suggest to the user to use admin://",
+                          FALSE,
+                          (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+
+  properties [PROP_HAD_ERROR] =
+    g_param_spec_boolean ("had-error",
+                          "Had Error",
+                          "If there was an error with the document",
+                          FALSE,
+                          (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+
   properties [PROP_BUSY] =
     g_param_spec_boolean ("busy",
                           "Busy",
@@ -543,15 +704,33 @@ editor_document_class_init (EditorDocumentClass *klass)
                          (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 
   g_object_class_install_properties (object_class, N_PROPS, properties);
+
+  /**
+   * EditorDocument::save:
+   * @self: a #EditorDocument
+   *
+   * The "save" signal is emitted right before the document
+   * is about to be saved.
+   */
+  signals [SAVE] = g_signal_new ("save",
+                                 G_TYPE_FROM_CLASS (klass),
+                                 G_SIGNAL_RUN_LAST,
+                                 0,
+                                 NULL, NULL,
+                                 NULL,
+                                 G_TYPE_NONE, 0);
 }
 
 static void
 editor_document_init (EditorDocument *self)
 {
+  g_autoptr(EditorSpellChecker) spell_checker = editor_spell_checker_new (NULL, NULL);
+
   self->newline_type = GTK_SOURCE_NEWLINE_TYPE_DEFAULT;
   self->file = gtk_source_file_new ();
   self->draft_id = g_uuid_string_random ();
-  self->spell_checker = editor_spell_checker_new (NULL, NULL);
+
+  editor_document_set_spell_checker (self, spell_checker);
 
   self->monitor = editor_buffer_monitor_new ();
   g_signal_connect_object (self->monitor,
@@ -724,6 +903,8 @@ _editor_document_save_draft_async (EditorDocument      *self,
       return;
     }
 
+  editor_document_emit_save (self);
+
   self->needs_autosave = FALSE;
 
   /* First tell the session to track this draft */
@@ -747,6 +928,17 @@ _editor_document_save_draft_async (EditorDocument      *self,
   if (self->encoding != NULL)
     gtk_source_file_saver_set_encoding (saver, self->encoding);
 
+  /* If there are no modifications, then delete any backing file
+   * for the draft so we don't risk reloading it.
+   */
+  if (!gtk_text_buffer_get_modified (GTK_TEXT_BUFFER (self)) &&
+      editor_document_get_file (self) != NULL)
+    {
+      g_file_delete (draft_file, NULL, NULL);
+      g_task_return_boolean (task, TRUE);
+      return;
+    }
+
   /* TODO: Probably want to make this async. We can just create an
    * async variant in editor-utils.c for this.
    */
@@ -901,7 +1093,7 @@ editor_document_save_cb (GObject      *object,
   g_file_delete_async (draft, G_PRIORITY_DEFAULT, NULL, delete_draft_cb, NULL);
 
   info = g_file_info_new ();
-  g_file_info_set_attribute_string (info, METATDATA_CURSOR, save->position);
+  g_file_info_set_attribute_string (info, METADATA_CURSOR, save->position);
   g_file_set_attributes_async (file,
                                info,
                                G_FILE_QUERY_INFO_NONE,
@@ -928,6 +1120,8 @@ _editor_document_save_async (EditorDocument      *self,
   g_return_if_fail (!cancellable || G_IS_CANCELLABLE (cancellable));
   g_return_if_fail (self->draft_id != NULL);
 
+  editor_document_emit_save (self);
+
   insert = gtk_text_buffer_get_insert (GTK_TEXT_BUFFER (self));
   gtk_text_buffer_get_iter_at_mark (GTK_TEXT_BUFFER (self), &iter, insert);
 
@@ -1077,10 +1271,12 @@ editor_document_query_info_cb (GObject      *object,
   g_autoptr(GTask) task = user_data;
   g_autofree gchar *filename = NULL;
   GtkSourceLanguageManager *lm;
-  GtkSourceLanguage *language;
+  GtkSourceLanguage *language = NULL;
   EditorDocument *self;
   const gchar *content_type;
   const gchar *position;
+  const gchar *spelling_language;
+  const gchar *syntax;
   gboolean readonly;
   gboolean is_modified;
   guint line = 0;
@@ -1107,11 +1303,20 @@ editor_document_query_info_cb (GObject      *object,
 
   readonly = g_file_info_get_attribute_boolean (info, G_FILE_ATTRIBUTE_FILESYSTEM_READONLY);
   content_type = g_file_info_get_attribute_string (info, G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE);
-  position = g_file_info_get_attribute_string (info, METATDATA_CURSOR);
+  position = g_file_info_get_attribute_string (info, METADATA_CURSOR);
+  spelling_language = g_file_info_get_attribute_string (info, METADATA_SPELLING);
+  syntax = g_file_info_get_attribute_string (info, METADATA_SYNTAX);
 
   editor_document_set_readonly (self, readonly);
 
-  language = gtk_source_language_manager_guess_language (lm, filename, content_type);
+  /* Use existing language from extended attribute if it is available, otherwise
+   * fallback to guessing the syntax from the content-type.
+   */
+  if (syntax != NULL && syntax[0] != 0)
+    language = gtk_source_language_manager_get_language (lm, syntax);
+  if (language == NULL)
+    language = guess_language (lm, filename, content_type);
+
   gtk_source_buffer_set_language (GTK_SOURCE_BUFFER (self), language);
   gtk_source_buffer_set_highlight_syntax (GTK_SOURCE_BUFFER (self), language != NULL);
 
@@ -1129,6 +1334,10 @@ editor_document_query_info_cb (GObject      *object,
       gtk_text_buffer_select_range (GTK_TEXT_BUFFER (self), &iter, &iter);
     }
 
+  /* Apply metadata for spelling language */
+  if (spelling_language != NULL && self->spell_checker != NULL)
+    editor_spell_checker_set_language (self->spell_checker, spelling_language);
+
   load = g_task_get_task_data (task);
 
   is_modified = load->has_draft;
@@ -1175,6 +1384,7 @@ editor_document_load_cb (GObject      *object,
   if (!gtk_source_file_loader_load_finish (loader, result, &error))
     {
       g_warning ("Failed to load file: %s", error->message);
+      editor_document_track_error (self, error);
       g_task_return_error (task, g_steal_pointer (&error));
       _editor_document_unmark_busy (self);
       return;
@@ -1200,7 +1410,9 @@ editor_document_load_cb (GObject      *object,
       g_file_query_info_async (file,
                                G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE","
                                G_FILE_ATTRIBUTE_FILESYSTEM_READONLY","
-                               METATDATA_CURSOR,
+                               METADATA_CURSOR","
+                               METADATA_SYNTAX","
+                               METADATA_SPELLING,
                                G_FILE_QUERY_INFO_NONE,
                                G_PRIORITY_DEFAULT,
                                g_task_get_cancellable (task),
@@ -1332,12 +1544,20 @@ editor_document_load_file_info_cb (GObject      *object,
           load->has_file = TRUE;
           load->content_type = NULL;
           load->modified_at = 0;
+          g_clear_error (&error);
         }
       else if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_NOT_FOUND))
         {
           load->has_file = FALSE;
           load->content_type = NULL;
           load->modified_at = 0;
+          g_clear_error (&error);
+        }
+      else if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED))
+        {
+          load->has_file = TRUE;
+          load->content_type = NULL;
+          load->modified_at = 0;
         }
       else
         {
@@ -1352,10 +1572,20 @@ editor_document_load_file_info_cb (GObject      *object,
       load->modified_at = g_file_info_get_attribute_uint64 (info, G_FILE_ATTRIBUTE_TIME_MODIFIED);
     }
 
+  /* XXX: We might need to be more careful about how we call this multiple times */
+  editor_document_track_error (self, error);
+
   load->n_active--;
 
   if (load->n_active == 0)
-    editor_document_do_load (self, task, load);
+    {
+      if (error != NULL)
+        g_task_return_error (task, g_steal_pointer (&error));
+      else if (self->last_error)
+        g_task_return_error (task, g_error_copy (self->last_error));
+      else
+        editor_document_do_load (self, task, load);
+    }
 }
 
 static void
@@ -1409,17 +1639,6 @@ editor_document_load_file_mount_cb (GObject      *object,
                            g_object_ref (task));
 }
 
-static void
-editor_document_load_completed_cb (EditorDocument *self,
-                                   GParamSpec     *pspec,
-                                   GTask          *task)
-{
-  g_assert (EDITOR_IS_DOCUMENT (self));
-  g_assert (G_IS_TASK (task));
-
-  self->loading = FALSE;
-}
-
 void
 _editor_document_load_async (EditorDocument      *self,
                              EditorWindow        *window,
@@ -1456,12 +1675,6 @@ _editor_document_load_async (EditorDocument      *self,
   g_task_set_source_tag (task, _editor_document_load_async);
   g_task_set_task_data (task, load, (GDestroyNotify) load_free);
 
-  g_signal_connect_object (task,
-                           "notify::completed",
-                           G_CALLBACK (editor_document_load_completed_cb),
-                           self,
-                           G_CONNECT_SWAPPED);
-
   editor_document_set_busy_progress (self, 0, 2, .25);
 
   g_signal_connect_object (task,
@@ -1545,7 +1758,7 @@ editor_document_guess_language_query_cb (GObject      *object,
   content_type = g_file_info_get_attribute_string (info, G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE);
   filename = g_file_get_basename (file);
   lm = gtk_source_language_manager_get_default ();
-  language = gtk_source_language_manager_guess_language (lm, filename, content_type);
+  language = guess_language (lm, filename, content_type);
   self = g_task_get_source_object (task);
 
   g_assert (EDITOR_IS_DOCUMENT (self));
@@ -1588,6 +1801,11 @@ _editor_document_guess_language_async (EditorDocument      *self,
                              G_IO_ERROR,
                              G_IO_ERROR_INVALID_FILENAME,
                              "File has not been saved, cannot guess content-type");
+  else if (self->load_failed)
+    g_task_return_new_error (task,
+                             G_IO_ERROR,
+                             G_IO_ERROR_CANCELLED,
+                             "Cannot query file as load failed.");
   else
     g_file_query_info_async (file,
                              G_FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
@@ -1655,11 +1873,14 @@ editor_document_dup_title (EditorDocument *self)
   g_return_val_if_fail (EDITOR_IS_DOCUMENT (self), NULL);
 
   file = editor_document_get_file (self);
+  str = g_string_new (NULL);
 
   if (file != NULL)
-    return g_file_get_basename (file);
-
-  str = g_string_new (NULL);
+    {
+      g_autofree char *base = g_file_get_basename (file);
+      g_string_append (str, base);
+      goto handle_suffix;
+    }
 
   gtk_text_buffer_get_start_iter (GTK_TEXT_BUFFER (self), &iter);
 
@@ -1709,6 +1930,8 @@ editor_document_dup_title (EditorDocument *self)
   if (str->len > 0 && str->str[str->len-1] == ' ')
     g_string_truncate (str, str->len - 1);
 
+handle_suffix:
+
   if (self->readonly)
     {
       g_string_append_c (str, ' ');
@@ -1836,11 +2059,28 @@ editor_document_set_spell_checker (EditorDocument     *self,
   g_return_if_fail (EDITOR_IS_DOCUMENT (self));
   g_return_if_fail (!spell_checker || EDITOR_IS_SPELL_CHECKER (spell_checker));
 
-  if (g_set_object (&self->spell_checker, spell_checker))
+  if (spell_checker == self->spell_checker)
+    return;
+
+  if (self->spell_checker != NULL)
     {
-      editor_text_buffer_spell_adapter_set_checker (self->spell_adapter, spell_checker);
-      g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_SPELL_CHECKER]);
+      g_signal_handlers_disconnect_by_func (self->spell_checker,
+                                            G_CALLBACK (on_spelling_language_changed_cb),
+                                            self);
+      g_clear_object (&self->spell_checker);
     }
+
+  if (spell_checker != NULL)
+    {
+      self->spell_checker = g_object_ref (spell_checker);
+      g_signal_connect_object (self->spell_checker,
+                               "notify::language",
+                               G_CALLBACK (on_spelling_language_changed_cb),
+                               self,
+                               G_CONNECT_SWAPPED);
+    }
+
+  g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_SPELL_CHECKER]);
 }
 
 void
@@ -1923,3 +2163,36 @@ _editor_document_get_spelling_tag (EditorDocument *self)
 
   return editor_text_buffer_spell_adapter_get_tag (self->spell_adapter);
 }
+
+void
+_editor_document_use_admin (EditorDocument *self)
+{
+  GFile *file;
+  g_autofree char *uri = NULL;
+  g_autofree char *admin_uri = NULL;
+  g_autoptr(GFile) admin_file = NULL;
+
+  g_return_if_fail (EDITOR_IS_DOCUMENT (self));
+
+  if (!(file = editor_document_get_file (self)))
+    return;
+
+  uri = g_file_get_uri (file);
+  if (!g_str_has_prefix (uri, "file:///"))
+    return;
+
+  admin_uri = g_strdup_printf ("admin://%s", g_file_get_path (file));
+  admin_file = g_file_new_for_uri (admin_uri);
+
+  gtk_source_file_set_location (self->file, admin_file);
+
+  g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_FILE]);
+}
+
+gboolean
+_editor_document_had_error (EditorDocument *self)
+{
+  g_return_val_if_fail (EDITOR_IS_DOCUMENT (self), FALSE);
+
+  return self->load_failed;
+}
