@@ -24,7 +24,9 @@
 
 #include <adwaita.h>
 #include <glib/gi18n.h>
+#include <unistd.h>
 
+#include "editor-application-private.h"
 #include "editor-info-bar-private.h"
 #include "editor-page-private.h"
 #include "editor-sidebar-model-private.h"
@@ -37,10 +39,14 @@ enum {
   PROP_BUSY,
   PROP_CAN_SAVE,
   PROP_DOCUMENT,
+  PROP_INDICATOR,
   PROP_IS_MODIFIED,
+  PROP_LANGUAGE_NAME,
+  PROP_POSITION_LABEL,
+  PROP_SETTINGS,
   PROP_SUBTITLE,
   PROP_TITLE,
-  PROP_POSITION_LABEL,
+  PROP_ZOOM_LABEL,
   N_PROPS
 };
 
@@ -48,6 +54,65 @@ G_DEFINE_TYPE (EditorPage, editor_page, GTK_TYPE_WIDGET)
 
 static GParamSpec *properties [N_PROPS];
 
+static void
+on_draw_spaces_changed (EditorPage  *self,
+                        const gchar *key,
+                        GSettings   *settings)
+{
+  GtkSourceSpaceLocationFlags location_flags = GTK_SOURCE_SPACE_LOCATION_NONE;
+  GtkSourceSpaceTypeFlags type_flags = GTK_SOURCE_SPACE_TYPE_NONE;
+  GtkSourceView *source_view;
+  GtkSourceSpaceDrawer *drawer;
+  guint flags;
+
+  g_assert (EDITOR_IS_PAGE (self));
+  g_assert (g_strcmp0 (key, "draw-spaces") == 0);
+  g_assert (G_IS_SETTINGS (settings));
+
+  source_view = GTK_SOURCE_VIEW (self->view);
+  drawer = gtk_source_view_get_space_drawer (source_view);
+  flags = g_settings_get_flags (settings, "draw-spaces");
+
+  if (flags == 0)
+    {
+      gtk_source_space_drawer_set_enable_matrix (drawer, FALSE);
+      return;
+    }
+
+  /* Reset the matrix before setting it */
+  gtk_source_space_drawer_set_types_for_locations (drawer, GTK_SOURCE_SPACE_LOCATION_ALL, GTK_SOURCE_SPACE_TYPE_NONE);
+
+  if (flags & 1)
+    type_flags |= GTK_SOURCE_SPACE_TYPE_SPACE;
+
+  if (flags & 2)
+    type_flags |= GTK_SOURCE_SPACE_TYPE_TAB;
+
+  if (flags & 4)
+    {
+      gtk_source_space_drawer_set_types_for_locations (drawer, GTK_SOURCE_SPACE_LOCATION_ALL, GTK_SOURCE_SPACE_TYPE_NEWLINE);
+      type_flags |= GTK_SOURCE_SPACE_TYPE_NEWLINE;
+    }
+
+  if (flags & 8)
+    type_flags |= GTK_SOURCE_SPACE_TYPE_NBSP;
+
+  if (flags & 16)
+    location_flags |= GTK_SOURCE_SPACE_LOCATION_LEADING;
+
+  if (flags & 32)
+    location_flags |= GTK_SOURCE_SPACE_LOCATION_INSIDE_TEXT;
+
+  if (flags & 64)
+    location_flags |= GTK_SOURCE_SPACE_LOCATION_TRAILING;
+
+  if (type_flags > 0 && location_flags == 0)
+    location_flags |= GTK_SOURCE_SPACE_LOCATION_ALL;
+
+  gtk_source_space_drawer_set_enable_matrix (drawer, TRUE);
+  gtk_source_space_drawer_set_types_for_locations (drawer, location_flags, type_flags);
+}
+
 static void
 editor_page_set_settings (EditorPage         *self,
                           EditorPageSettings *settings)
@@ -135,6 +200,7 @@ editor_page_document_notify_file_cb (EditorPage     *self,
   g_assert (EDITOR_IS_PAGE (self));
   g_assert (EDITOR_IS_DOCUMENT (document));
 
+  g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_INDICATOR]);
   g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_TITLE]);
   g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_SUBTITLE]);
 }
@@ -151,6 +217,8 @@ editor_page_document_notify_language_cb (EditorPage     *self,
 
   settings = editor_page_settings_new_for_document (document);
   editor_page_set_settings (self, settings);
+
+  g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_LANGUAGE_NAME]);
 }
 
 static void
@@ -246,7 +314,7 @@ boolean_to_left_margin (GBinding     *binding,
                         gpointer      user_data)
 {
   if (g_value_get_boolean (from_value))
-    g_value_set_int (to_value, 0);
+    g_value_set_int (to_value, 6);
   else
     g_value_set_int (to_value, 12);
   return TRUE;
@@ -268,10 +336,26 @@ font_desc_from_string (GBinding     *binding,
   return TRUE;
 }
 
+static void
+on_settings_changed_cb (EditorPage         *self,
+                        GParamSpec         *pspec,
+                        EditorBindingGroup *group)
+{
+  EditorPageSettings *settings;
+
+  g_assert (EDITOR_IS_PAGE (self));
+  g_assert (EDITOR_IS_BINDING_GROUP (group));
+
+  settings = EDITOR_PAGE_SETTINGS (editor_binding_group_get_source (group));
+
+  _editor_page_actions_bind_settings (self, settings);
+}
+
 static void
 editor_page_constructed (GObject *object)
 {
   EditorPage *self = (EditorPage *)object;
+  EditorApplication *app;
 
   g_assert (EDITOR_IS_PAGE (self));
 
@@ -279,6 +363,12 @@ editor_page_constructed (GObject *object)
 
   self->settings_bindings = editor_binding_group_new ();
 
+  g_signal_connect_object (self->settings_bindings,
+                           "notify::source",
+                           G_CALLBACK (on_settings_changed_cb),
+                           self,
+                           G_CONNECT_SWAPPED);
+
   editor_binding_group_bind_full (self->settings_bindings, "custom-font",
                                   self->view, "font-desc",
                                   G_BINDING_SYNC_CREATE,
@@ -314,6 +404,12 @@ editor_page_constructed (GObject *object)
   editor_binding_group_bind (self->settings_bindings, "tab-width",
                              self->view, "tab-width",
                              G_BINDING_SYNC_CREATE);
+  editor_binding_group_bind (self->settings_bindings, "indent-width",
+                             self->view, "indent-width",
+                             G_BINDING_SYNC_CREATE);
+  editor_binding_group_bind (self->settings_bindings, "auto-indent",
+                             self->view, "auto-indent",
+                             G_BINDING_SYNC_CREATE);
   editor_binding_group_bind_full (self->settings_bindings, "wrap-text",
                                   self->view, "wrap-mode",
                                   G_BINDING_SYNC_CREATE,
@@ -334,6 +430,24 @@ editor_page_constructed (GObject *object)
 
   editor_page_document_notify_busy_cb (self, NULL, self->document);
   editor_page_document_notify_language_cb (self, NULL, self->document);
+
+  /* Setup spaces drawing */
+  app = EDITOR_APPLICATION (g_application_get_default ());
+  g_signal_connect_object (app->settings,
+                           "changed::draw-spaces",
+                           G_CALLBACK (on_draw_spaces_changed),
+                           self,
+                           G_CONNECT_SWAPPED);
+  on_draw_spaces_changed (self, "draw-spaces", app->settings);
+
+  g_settings_bind (app->settings, "enable-snippets",
+                   self->view, "enable-snippets",
+                   G_SETTINGS_BIND_GET);
+  g_settings_bind (app->settings, "line-height",
+                   self->view, "line-height",
+                   G_SETTINGS_BIND_GET);
+
+  _editor_page_vim_init (self);
 }
 
 static gboolean
@@ -346,18 +460,63 @@ editor_page_drop_target_drop (EditorPage     *self,
   g_assert (EDITOR_IS_PAGE (self));
   g_assert (GTK_IS_DROP_TARGET (dest));
 
-  if (G_VALUE_HOLDS (value, G_TYPE_FILE))
+  if (G_VALUE_HOLDS (value, GDK_TYPE_FILE_LIST))
     {
-      GFile *file = g_value_get_object (value);
       EditorSession *session = editor_application_get_session (EDITOR_APPLICATION_DEFAULT);
       EditorWindow *window = EDITOR_WINDOW (gtk_widget_get_root (GTK_WIDGET (self)));
+      GSList *list = g_value_get_boxed (value);
 
-      editor_session_open (session, window, file, NULL);
+      for (const GSList *iter = list; iter; iter = iter->next)
+        {
+          GFile *file = iter->data;
+          g_assert (G_IS_FILE (file));
+          editor_session_open (session, window, file, NULL);
+        }
     }
 
   return FALSE;
 }
 
+static void
+goto_line_entry_activate_cb (EditorPage *self,
+                             GtkEntry   *entry)
+{
+  g_assert (EDITOR_IS_PAGE (self));
+  g_assert (GTK_IS_ENTRY (entry));
+
+  gtk_widget_activate_action (GTK_WIDGET (self), "page.goto-line", NULL);
+}
+
+static void
+goto_line_entry_insert_text_cb (EditorPage *self,
+                                const char *new_text,
+                                int         length,
+                                int        *position,
+                                GtkText    *entry)
+{
+  const char *text;
+  GString *str;
+
+  g_assert (EDITOR_IS_PAGE (self));
+  g_assert (position != NULL);
+  g_assert (GTK_IS_TEXT (entry));
+
+  text = gtk_editable_get_text (GTK_EDITABLE (entry));
+  str = g_string_new (text);
+  g_string_insert_len (str, *position, new_text, length);
+
+  for (const char *c = str->str; *c; c = g_utf8_next_char (c))
+    {
+      gunichar ch = g_utf8_get_char (c);
+
+      if (ch == ':' || (ch >= '0' && ch <= '9'))
+        continue;
+
+      g_signal_stop_emission_by_name (entry, "insert-text");
+      return;
+    }
+}
+
 static gboolean
 get_child_position_cb (GtkOverlay    *overlay,
                        GtkWidget     *child,
@@ -365,6 +524,7 @@ get_child_position_cb (GtkOverlay    *overlay,
                        EditorPage    *self)
 {
   GtkRequisition min, nat, map_min;
+  GtkAllocation ov_alloc;
 
   g_assert (GTK_IS_OVERLAY (overlay));
   g_assert (GTK_IS_WIDGET (child));
@@ -375,12 +535,13 @@ get_child_position_cb (GtkOverlay    *overlay,
       !gtk_widget_get_visible (GTK_WIDGET (self->map)))
     return FALSE;
 
+  gtk_widget_get_allocation (GTK_WIDGET (overlay), &ov_alloc);
   gtk_widget_get_preferred_size (child, &min, &nat);
   gtk_widget_get_preferred_size (GTK_WIDGET (self->map), &map_min, NULL);
 
-  if (nat.width + map_min.width <= alloc->width)
+  if (nat.width + map_min.width <= ov_alloc.width)
     {
-      alloc->x += alloc->width;
+      alloc->x = ov_alloc.width;
       alloc->x -= map_min.width;
       alloc->x -= nat.width;
       alloc->width = nat.width;
@@ -390,7 +551,7 @@ get_child_position_cb (GtkOverlay    *overlay,
   else
     {
       alloc->x = 0;
-      alloc->width -= map_min.width;
+      alloc->width = min.width;
       alloc->y = 0;
       alloc->height = min.height;
     }
@@ -398,6 +559,58 @@ get_child_position_cb (GtkOverlay    *overlay,
   return TRUE;
 }
 
+static void
+font_scale_changed_cb (EditorPage       *self,
+                       GParamSpec       *pspec,
+                       EditorSourceView *view)
+{
+  g_assert (EDITOR_IS_PAGE (self));
+  g_assert (EDITOR_IS_SOURCE_VIEW (view));
+
+  g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_ZOOM_LABEL]);
+}
+
+const char *
+editor_page_get_language_name (EditorPage *self)
+{
+  GtkSourceLanguage *language;
+  EditorDocument *document;
+
+  g_return_val_if_fail (EDITOR_IS_PAGE (self), NULL);
+
+  document = editor_page_get_document (self);
+  language = gtk_source_buffer_get_language (GTK_SOURCE_BUFFER (document));
+
+  if (language != NULL)
+    return gtk_source_language_get_name (language);
+
+  return NULL;
+}
+
+static GIcon *
+editor_page_get_indicator (EditorPage *self)
+{
+  static GIcon *icon;
+  g_autofree char *uri = NULL;
+  EditorDocument *document;
+  GFile *file;
+
+  document = editor_page_get_document (self);
+  file = editor_document_get_file (document);
+
+  if (file == NULL)
+    return NULL;
+
+  uri = g_file_get_uri (file);
+  if (!g_str_has_prefix (uri, "admin://"))
+    return NULL;
+
+  if (icon == NULL)
+    icon = g_themed_icon_new ("document-admin-symbolic");
+
+  return icon;
+}
+
 static void
 editor_page_dispose (GObject *object)
 {
@@ -445,10 +658,18 @@ editor_page_get_property (GObject    *object,
       g_value_set_object (value, editor_page_get_document (self));
       break;
 
+    case PROP_INDICATOR:
+      g_value_set_object (value, editor_page_get_indicator (self));
+      break;
+
     case PROP_IS_MODIFIED:
       g_value_set_boolean (value, editor_page_get_is_modified (self));
       break;
 
+    case PROP_LANGUAGE_NAME:
+      g_value_set_string (value, editor_page_get_language_name (self));
+      break;
+
     case PROP_TITLE:
       g_value_take_string (value, editor_page_dup_title (self));
       break;
@@ -461,6 +682,14 @@ editor_page_get_property (GObject    *object,
       g_value_take_string (value, editor_page_dup_position_label (self));
       break;
 
+    case PROP_SETTINGS:
+      g_value_set_object (value, self->settings);
+      break;
+
+    case PROP_ZOOM_LABEL:
+      g_value_take_string (value, _editor_page_get_zoom_label (self));
+      break;
+
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
     }
@@ -518,6 +747,20 @@ editor_page_class_init (EditorPageClass *klass)
                          EDITOR_TYPE_DOCUMENT,
                          (G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS));
 
+  properties [PROP_INDICATOR] =
+    g_param_spec_object ("indicator",
+                         "Indicator",
+                         "The document indicator if any",
+                         G_TYPE_ICON,
+                         (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+
+  properties [PROP_LANGUAGE_NAME] =
+    g_param_spec_string ("language-name",
+                         "Language Name",
+                         "Language Name",
+                         NULL,
+                         (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+
   properties [PROP_TITLE] =
     g_param_spec_string ("title",
                          "Title",
@@ -546,6 +789,20 @@ editor_page_class_init (EditorPageClass *klass)
                          NULL,
                          (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 
+  properties [PROP_SETTINGS] =
+    g_param_spec_object ("settings",
+                         "Settings",
+                         "Settings",
+                         EDITOR_TYPE_PAGE_SETTINGS,
+                         (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+
+  properties [PROP_ZOOM_LABEL] =
+    g_param_spec_string ("zoom-label",
+                         "Zoom Label",
+                         "Zoom Label",
+                         NULL,
+                         (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+
   g_object_class_install_properties (object_class, N_PROPS, properties);
 
   _editor_page_class_actions_init (klass);
@@ -554,6 +811,9 @@ editor_page_class_init (EditorPageClass *klass)
   gtk_widget_class_set_css_name (widget_class, "page");
   gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/TextEditor/ui/editor-page.ui");
   gtk_widget_class_bind_template_child (widget_class, EditorPage, box);
+  gtk_widget_class_bind_template_child (widget_class, EditorPage, goto_line_entry);
+  gtk_widget_class_bind_template_child (widget_class, EditorPage, goto_line_revealer);
+  gtk_widget_class_bind_template_child (widget_class, EditorPage, indent_menu);
   gtk_widget_class_bind_template_child (widget_class, EditorPage, infobar);
   gtk_widget_class_bind_template_child (widget_class, EditorPage, map);
   gtk_widget_class_bind_template_child (widget_class, EditorPage, overlay);
@@ -563,8 +823,11 @@ editor_page_class_init (EditorPageClass *klass)
   gtk_widget_class_bind_template_child (widget_class, EditorPage, search_revealer);
   gtk_widget_class_bind_template_child (widget_class, EditorPage, view);
   gtk_widget_class_bind_template_callback (widget_class, get_child_position_cb);
+  gtk_widget_class_bind_template_callback (widget_class, goto_line_entry_activate_cb);
+  gtk_widget_class_bind_template_callback (widget_class, font_scale_changed_cb);
 
   gtk_widget_class_add_binding_action (widget_class, GDK_KEY_Escape, 0, "search.hide", NULL);
+  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_i, GDK_CONTROL_MASK, "page.show-goto-line", NULL);
 
   g_type_ensure (EDITOR_TYPE_INFO_BAR);
   g_type_ensure (EDITOR_TYPE_SEARCH_BAR);
@@ -578,9 +841,29 @@ editor_page_init (EditorPage *self)
 
   gtk_widget_init_template (GTK_WIDGET (self));
 
+  editor_source_view_prepend_extra_menu (EDITOR_SOURCE_VIEW (self->view),
+                                         G_MENU_MODEL (self->indent_menu));
+  _editor_revealer_auto_hide (self->search_revealer);
+  _editor_revealer_auto_hide (self->goto_line_revealer);
+
+  /* Work around https://gitlab.gnome.org/GNOME/gtk/-/issues/4315
+   * by connecting to the GtkText to intercept insert-text() emission.
+   */
+  for (GtkWidget *child = gtk_widget_get_first_child (GTK_WIDGET (self->goto_line_entry));
+       child;
+       child = gtk_widget_get_next_sibling (child))
+    {
+      if (GTK_IS_TEXT (child))
+        g_signal_connect_object (child,
+                                 "insert-text",
+                                 G_CALLBACK (goto_line_entry_insert_text_cb),
+                                 self,
+                                 G_CONNECT_SWAPPED);
+    }
+
   g_object_bind_property (self, "document", self->infobar, "document", 0);
 
-  dest = gtk_drop_target_new (G_TYPE_FILE, GDK_ACTION_COPY);
+  dest = gtk_drop_target_new (GDK_TYPE_FILE_LIST, GDK_ACTION_COPY);
   g_signal_connect_object (dest,
                            "drop",
                            G_CALLBACK (editor_page_drop_target_drop),
@@ -709,21 +992,43 @@ editor_page_dup_title (EditorPage *self)
 gchar *
 editor_page_dup_subtitle (EditorPage *self)
 {
+  static char *docportal = NULL;
   g_autoptr(GFile) dir = NULL;
+  const char *peek;
   GFile *file;
 
   g_return_val_if_fail (EDITOR_IS_PAGE (self), NULL);
   g_return_val_if_fail (EDITOR_IS_DOCUMENT (self->document), NULL);
 
+  if (docportal == NULL)
+    docportal = g_strdup_printf ("/run/user/%u/doc/", getuid ());
+
   file = editor_document_get_file (self->document);
 
   if (file == NULL || !(dir = g_file_get_parent (file)))
     return g_strdup (_("Draft"));
 
   if (!g_file_is_native (dir))
-    return g_file_get_uri (dir);
+    {
+      g_autofree char *uri = g_file_get_uri (dir);
 
-  return _editor_path_collapse (g_file_peek_path (dir));
+      if (g_str_has_prefix (uri, "admin:///"))
+        {
+          const char *suffix = uri + strlen ("admin://");
+          g_autofree char *path = _editor_path_collapse (suffix);
+
+          /* translators: %s is replaced with the path on the filesystem */
+          return g_strdup_printf (_("%s (Administrator)"), path);
+        }
+
+      return g_steal_pointer (&uri);
+    }
+
+  peek = g_file_peek_path (dir);
+  if (g_str_has_prefix (peek, docportal))
+    return g_strdup (_("Document Portal"));
+
+  return _editor_path_collapse (peek);
 }
 
 void
@@ -790,7 +1095,7 @@ _editor_page_save (EditorPage *self)
 
   if (editor_document_get_file (self->document) == NULL)
     {
-      _editor_page_save_as (self);
+      _editor_page_save_as (self, NULL);
       return;
     }
 
@@ -816,6 +1121,14 @@ editor_page_save_as_cb (EditorPage           *self,
       if ((dest = gtk_file_chooser_get_file (GTK_FILE_CHOOSER (native))))
         {
           GtkSourceNewlineType crlf = _editor_file_chooser_get_line_ending (GTK_FILE_CHOOSER (native));
+          g_autoptr(GFile) directory = gtk_file_chooser_get_current_folder (GTK_FILE_CHOOSER (native));
+
+          if (directory != NULL)
+            {
+              g_autoptr(GSettings) settings = g_settings_new ("org.gnome.TextEditor");
+              g_autofree char *uri = g_file_get_uri (directory);
+              g_settings_set_string (settings, "last-save-directory", uri);
+            }
 
           _editor_document_set_newline_type (self->document, crlf);
           _editor_document_save_async (self->document,
@@ -830,13 +1143,22 @@ editor_page_save_as_cb (EditorPage           *self,
 }
 
 void
-_editor_page_save_as (EditorPage *self)
+_editor_page_save_as (EditorPage *self,
+                      const char *filename)
 {
+  g_autoptr(GSettings) settings = NULL;
+  g_autoptr(GFile) directory = NULL;
   GtkFileChooserNative *native;
   EditorWindow *window;
+  g_autofree char *uri = NULL;
 
   g_return_if_fail (EDITOR_IS_PAGE (self));
 
+  settings = g_settings_new ("org.gnome.TextEditor");
+  uri = g_settings_get_string (settings, "last-save-directory");
+  if (uri && uri[0])
+    directory = g_file_new_for_uri (uri);
+
   _editor_page_raise (self);
 
   window = _editor_page_get_window (self);
@@ -850,6 +1172,28 @@ _editor_page_save_as (EditorPage *self)
   _editor_file_chooser_add_line_endings (GTK_FILE_CHOOSER (native),
                                          _editor_document_get_newline_type (self->document));
 
+  if (filename != NULL)
+    {
+      GFile *file = editor_document_get_file (self->document);
+      g_autoptr(GFile) selected = NULL;
+
+      if (file != NULL && !g_path_is_absolute (filename))
+        {
+          g_autoptr(GFile) parent = g_file_get_parent (file);
+          selected = g_file_get_child (parent, filename);
+        }
+      else
+        {
+          selected = g_file_new_for_path (filename);
+        }
+
+      gtk_file_chooser_set_file (GTK_FILE_CHOOSER (native), selected, NULL);
+    }
+  else if (directory != NULL)
+    {
+      gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (native), directory, NULL);
+    }
+
   g_signal_connect_object (native,
                            "response",
                            G_CALLBACK (editor_page_save_as_cb),
@@ -1037,6 +1381,7 @@ _editor_page_set_search_visible (EditorPage          *self,
       _editor_search_bar_detach (self->search_bar);
     }
 
+  gtk_revealer_set_reveal_child (self->goto_line_revealer, FALSE);
   gtk_revealer_set_reveal_child (self->search_revealer, search_visible);
 
   if (search_visible)
@@ -1079,19 +1424,21 @@ _editor_page_scroll_to_insert (EditorPage *self)
 }
 
 void
-_editor_page_move_next_search (EditorPage *self)
+_editor_page_move_next_search (EditorPage *self,
+                               gboolean    hide_after_move)
 {
   g_return_if_fail (EDITOR_IS_PAGE (self));
 
-  _editor_search_bar_move_next (self->search_bar, FALSE);
+  _editor_search_bar_move_next (self->search_bar, hide_after_move);
 }
 
 void
-_editor_page_move_previous_search (EditorPage *self)
+_editor_page_move_previous_search (EditorPage *self,
+                                   gboolean    hide_after_move)
 {
   g_return_if_fail (EDITOR_IS_PAGE (self));
 
-  _editor_search_bar_move_previous (self->search_bar, FALSE);
+  _editor_search_bar_move_previous (self->search_bar, hide_after_move);
 }
 
 static void
@@ -1121,6 +1468,7 @@ editor_page_delete_draft_cb (GObject      *object,
   EditorSession *session = EDITOR_SESSION_DEFAULT;
   g_autoptr(GTask) task = user_data;
   g_autoptr(GError) error = NULL;
+  gboolean reload;
   EditorPage *self;
 
   g_assert (G_IS_FILE (file));
@@ -1128,6 +1476,7 @@ editor_page_delete_draft_cb (GObject      *object,
   g_assert (G_IS_TASK (task));
 
   self = g_task_get_source_object (task);
+  reload = GPOINTER_TO_INT (g_task_get_task_data (task));
 
   if (!g_file_delete_finish (file, result, &error))
     {
@@ -1149,6 +1498,11 @@ editor_page_delete_draft_cb (GObject      *object,
                                     _editor_document_get_draft_id (self->document));
       g_task_return_boolean (task, TRUE);
     }
+  else if (!reload)
+    {
+      editor_session_remove_page (session, self);
+      g_task_return_boolean (task, TRUE);
+    }
   else
     {
       _editor_document_load_async (self->document,
@@ -1160,21 +1514,27 @@ editor_page_delete_draft_cb (GObject      *object,
 }
 
 void
-_editor_page_discard_changes_async  (EditorPage          *self,
-                                     GCancellable        *cancellable,
-                                     GAsyncReadyCallback  callback,
-                                     gpointer             user_data)
+_editor_page_discard_changes_async (EditorPage          *self,
+                                    gboolean             reload,
+                                    GCancellable        *cancellable,
+                                    GAsyncReadyCallback  callback,
+                                    gpointer             user_data)
 {
   g_autoptr(GFile) draft_file = NULL;
   g_autoptr(GTask) task = NULL;
+  const char *draft_id;
 
   g_return_if_fail (EDITOR_IS_PAGE (self));
   g_return_if_fail (!cancellable || G_IS_CANCELLABLE (cancellable));
 
   task = g_task_new (self, cancellable, callback, user_data);
   g_task_set_source_tag (task, _editor_page_discard_changes_async);
+  g_task_set_task_data (task, GINT_TO_POINTER (reload), NULL);
 
   draft_file = _editor_document_get_draft_file (self->document);
+  draft_id = _editor_document_get_draft_id (self->document);
+
+  _editor_session_remove_draft (EDITOR_SESSION_DEFAULT, draft_id);
 
   g_file_delete_async (draft_file,
                        G_PRIORITY_DEFAULT,
@@ -1200,7 +1560,7 @@ _editor_page_discard_changes (EditorPage *self)
   g_return_if_fail (EDITOR_IS_PAGE (self));
 
   _editor_page_raise (self);
-  _editor_page_discard_changes_async (self, NULL, NULL, NULL);
+  _editor_page_discard_changes_async (self, TRUE, NULL, NULL, NULL);
 }
 
 void
@@ -1211,3 +1571,66 @@ editor_page_grab_focus (EditorPage *self)
   _editor_page_raise (self);
   gtk_widget_grab_focus (GTK_WIDGET (self->view));
 }
+
+void
+_editor_page_begin_move (EditorPage *self)
+{
+  g_return_if_fail (!self->moving);
+
+  g_object_ref (self);
+  self->moving = TRUE;
+}
+
+void
+_editor_page_end_move (EditorPage *self)
+{
+  g_return_if_fail (self->moving);
+
+  self->moving = FALSE;
+  g_object_unref (self);
+}
+
+char *
+_editor_page_get_zoom_label (EditorPage *self)
+{
+  double level;
+
+  g_return_val_if_fail (EDITOR_IS_PAGE (self), NULL);
+
+  level = editor_source_view_get_zoom_level (EDITOR_SOURCE_VIEW (self->view));
+
+  if (level == 1.0)
+    return g_strdup ("100%");
+
+  return g_strdup_printf ("%.0lf%%", level * 100.0);
+}
+
+void
+_editor_page_zoom_in (EditorPage *self)
+{
+  int font_scale = 0;
+
+  g_return_if_fail (EDITOR_IS_PAGE (self));
+
+  g_object_get (self->view, "font-scale", &font_scale, NULL);
+  g_object_set (self->view, "font-scale", font_scale + 1, NULL);
+}
+
+void
+_editor_page_zoom_out (EditorPage *self)
+{
+  int font_scale = 0;
+
+  g_return_if_fail (EDITOR_IS_PAGE (self));
+
+  g_object_get (self->view, "font-scale", &font_scale, NULL);
+  g_object_set (self->view, "font-scale", font_scale - 1, NULL);
+}
+
+void
+_editor_page_zoom_one (EditorPage *self)
+{
+  g_return_if_fail (EDITOR_IS_PAGE (self));
+
+  g_object_set (self->view, "font-scale", 0, NULL);
+}
