@@ -25,38 +25,9 @@
 #include "editor-enums.h"
 #include "editor-page-private.h"
 #include "editor-search-bar-private.h"
+#include "editor-search-entry-private.h"
 #include "editor-utils-private.h"
 
-struct _EditorSearchBar
-{
-  GtkWidget                parent_instance;
-
-  GtkSourceSearchContext  *context;
-  GtkSourceSearchSettings *settings;
-  GCancellable            *cancellable;
-
-  GtkGrid                 *grid;
-  GtkEntry                *search_entry;
-  GtkEntry                *replace_entry;
-  GtkButton               *replace_button;
-  GtkButton               *replace_all_button;
-  GtkCheckButton          *case_button;
-  GtkCheckButton          *regex_button;
-  GtkCheckButton          *word_button;
-  GtkToggleButton         *options_button;
-  GtkToggleButton         *replace_mode_button;
-  GtkBox                  *options_box;
-
-  guint                    offset_when_shown;
-
-  guint                    can_move : 1;
-  guint                    can_replace : 1;
-  guint                    can_replace_all : 1;
-  guint                    hide_after_move : 1;
-  guint                    scroll_to_first_match : 1;
-  guint                    jump_back_on_hide : 1;
-};
-
 G_DEFINE_TYPE (EditorSearchBar, editor_search_bar, GTK_TYPE_WIDGET)
 
 enum {
@@ -71,7 +42,6 @@ enum {
 enum {
   MOVE_NEXT_SEARCH,
   MOVE_PREVIOUS_SEARCH,
-  HIDE_SEARCH,
   N_SIGNALS
 };
 
@@ -84,6 +54,7 @@ update_properties (EditorSearchBar *self)
   gboolean can_move = _editor_search_bar_get_can_move (self);
   gboolean can_replace = _editor_search_bar_get_can_replace (self);
   gboolean can_replace_all = _editor_search_bar_get_can_replace_all (self);
+  int occurrence_position = -1;
 
   if (can_move != self->can_move)
     {
@@ -102,6 +73,17 @@ update_properties (EditorSearchBar *self)
       self->can_replace_all = can_replace_all;
       g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_CAN_REPLACE_ALL]);
     }
+
+  if (self->context != NULL)
+    {
+      GtkTextBuffer *buffer = GTK_TEXT_BUFFER (gtk_source_search_context_get_buffer (self->context));
+      GtkTextIter begin, end;
+
+      if (gtk_text_buffer_get_selection_bounds (buffer, &begin, &end))
+        occurrence_position = gtk_source_search_context_get_occurrence_position (self->context, &begin, &end);
+    }
+
+  editor_search_entry_set_occurrence_position (self->search_entry, occurrence_position);
 }
 
 static void
@@ -115,16 +97,6 @@ editor_search_bar_scroll_to_insert (EditorSearchBar *self)
     _editor_page_scroll_to_insert (EDITOR_PAGE (page));
 }
 
-static void
-editor_search_bar_search_activate_cb (EditorSearchBar *self,
-                                      GtkEntry        *entry)
-{
-  g_assert (EDITOR_IS_SEARCH_BAR (self));
-  g_assert (GTK_IS_ENTRY (entry));
-
-  _editor_search_bar_move_next (self, TRUE);
-}
-
 static void
 editor_search_bar_move_next_forward_cb (GObject      *object,
                                         GAsyncResult *result,
@@ -179,7 +151,7 @@ _editor_search_bar_move_next (EditorSearchBar *self,
 
   gtk_source_search_context_forward_async (self->context,
                                            &end,
-                                           self->cancellable,
+                                           NULL,
                                            editor_search_bar_move_next_forward_cb,
                                            g_object_ref (self));
 }
@@ -206,20 +178,12 @@ _editor_search_bar_move_previous (EditorSearchBar *self,
 
   gtk_source_search_context_backward_async (self->context,
                                             &begin,
-                                            self->cancellable,
+                                            NULL,
                                             /* XXX: fixme */
                                             editor_search_bar_move_next_forward_cb,
                                             g_object_ref (self));
 }
 
-static void
-editor_search_bar_hide_search_cb (EditorSearchBar *self)
-{
-  g_assert (EDITOR_IS_SEARCH_BAR (self));
-
-  gtk_widget_activate_action (GTK_WIDGET (self), "search.hide", NULL);
-}
-
 static gboolean
 text_to_search_text (GBinding     *binding,
                      const GValue *from_value,
@@ -230,7 +194,7 @@ text_to_search_text (GBinding     *binding,
   const gchar *str = g_value_get_string (from_value);
 
   if (!str || gtk_source_search_settings_get_regex_enabled (self->settings))
-    g_value_set_string (to_value, NULL);
+    g_value_set_string (to_value, str);
   else
     g_value_take_string (to_value, gtk_source_utils_unescape_search_text (str));
 
@@ -303,12 +267,12 @@ on_notify_replace_text_cb (EditorSearchBar *self,
 }
 
 static void
-on_notify_search_text_cb (EditorSearchBar *self,
-                          GParamSpec      *pspec,
-                          GtkEntry        *entry)
+on_notify_search_text_cb (EditorSearchBar   *self,
+                          GParamSpec        *pspec,
+                          EditorSearchEntry *entry)
 {
   g_assert (EDITOR_IS_SEARCH_BAR (self));
-  g_assert (GTK_IS_ENTRY (entry));
+  g_assert (EDITOR_IS_SEARCH_ENTRY (entry));
 
   self->scroll_to_first_match = TRUE;
 }
@@ -360,7 +324,6 @@ editor_search_bar_finalize (GObject *object)
 {
   EditorSearchBar *self = (EditorSearchBar *)object;
 
-  g_clear_object (&self->cancellable);
   g_clear_object (&self->context);
   g_clear_object (&self->settings);
 
@@ -445,7 +408,6 @@ editor_search_bar_class_init (EditorSearchBarClass *klass)
   gtk_widget_class_bind_template_child (widget_class, EditorSearchBar, replace_mode_button);
   gtk_widget_class_bind_template_child (widget_class, EditorSearchBar, search_entry);
   gtk_widget_class_bind_template_child (widget_class, EditorSearchBar, word_button);
-  gtk_widget_class_bind_template_callback (widget_class, editor_search_bar_search_activate_cb);
   gtk_widget_class_bind_template_callback (widget_class, on_search_key_pressed_cb);
 
   signals [MOVE_NEXT_SEARCH] =
@@ -466,24 +428,7 @@ editor_search_bar_class_init (EditorSearchBarClass *klass)
                                 NULL,
                                 G_TYPE_NONE, 1, G_TYPE_BOOLEAN);
 
-  signals [HIDE_SEARCH] =
-    g_signal_new_class_handler ("hide-search",
-                                G_TYPE_FROM_CLASS (klass),
-                                G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
-                                G_CALLBACK (editor_search_bar_hide_search_cb),
-                                NULL, NULL,
-                                NULL,
-                                G_TYPE_NONE, 0);
-
   gtk_widget_class_add_binding_action (widget_class, GDK_KEY_Escape, 0, "search.hide", NULL);
-  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_g, GDK_CONTROL_MASK|GDK_SHIFT_MASK, "search.move-previous-search", NULL);
-  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_g, GDK_CONTROL_MASK, "search.move-next-search", NULL);
-  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_Down, 0, "search.move-next-search", NULL);
-  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_Up, 0, "search.move-previous-search", NULL);
-  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_Return, 0, "search.move-next-search", NULL);
-  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_KP_Enter, 0, "search.move-next-search", NULL);
-  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_Return, GDK_SHIFT_MASK, "search.move-previous-search", NULL);
-  gtk_widget_class_add_binding_action (widget_class, GDK_KEY_KP_Enter, GDK_SHIFT_MASK, "search.move-previous-search", NULL);
 
   properties [PROP_MODE] =
     g_param_spec_enum ("mode",
@@ -515,6 +460,8 @@ editor_search_bar_class_init (EditorSearchBarClass *klass)
                           (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 
   g_object_class_install_properties (object_class, N_PROPS, properties);
+
+  g_type_ensure (EDITOR_TYPE_SEARCH_ENTRY);
 }
 
 static void
@@ -611,11 +558,15 @@ editor_search_bar_notify_occurrences_count_cb (EditorSearchBar        *self,
                                                GParamSpec             *pspec,
                                                GtkSourceSearchContext *context)
 {
+  guint occurrence_count;
+
   g_assert (EDITOR_IS_SEARCH_BAR (self));
   g_assert (GTK_SOURCE_IS_SEARCH_CONTEXT (context));
 
-  if (self->scroll_to_first_match &&
-      gtk_source_search_context_get_occurrences_count (context) > 0)
+  occurrence_count = gtk_source_search_context_get_occurrences_count (context);
+  editor_search_entry_set_occurrence_count (self->search_entry, occurrence_count);
+
+  if (self->scroll_to_first_match && occurrence_count > 0)
     scroll_to_first_match (self, context);
 
   update_properties (self);
@@ -637,26 +588,21 @@ _editor_search_bar_attach (EditorSearchBar *self,
 {
   GtkTextBuffer *buffer = (GtkTextBuffer *)document;
   GtkTextIter begin, end, insert;
-  const gchar *search;
 
   g_return_if_fail (EDITOR_IS_SEARCH_BAR (self));
 
-  if (self->context != NULL)
-    return;
-
   gtk_text_buffer_get_iter_at_mark (buffer, &insert, gtk_text_buffer_get_insert (buffer));
   self->offset_when_shown = gtk_text_iter_get_offset (&insert);
 
-  search = gtk_editable_get_text (GTK_EDITABLE (self->search_entry));
-
-  if ((search == NULL || g_strcmp0 (search, "") == 0) &&
-      gtk_text_buffer_get_selection_bounds (GTK_TEXT_BUFFER (document), &begin, &end))
+  if (gtk_text_buffer_get_selection_bounds (GTK_TEXT_BUFFER (document), &begin, &end))
     {
       g_autofree gchar *text = gtk_text_iter_get_slice (&begin, &end);
       gtk_editable_set_text (GTK_EDITABLE (self->search_entry), text);
     }
 
-  self->cancellable = g_cancellable_new ();
+  if (self->context != NULL)
+    return;
+
   self->context = gtk_source_search_context_new (GTK_SOURCE_BUFFER (document), self->settings);
 
   g_signal_connect_object (self->context,
@@ -695,9 +641,6 @@ _editor_search_bar_detach (EditorSearchBar *self)
       g_clear_object (&self->context);
     }
 
-  g_cancellable_cancel (self->cancellable);
-  g_clear_object (&self->cancellable);
-
   self->hide_after_move = FALSE;
   self->jump_back_on_hide = FALSE;
 }
@@ -756,9 +699,13 @@ _editor_search_bar_replace (EditorSearchBar *self)
   gtk_text_buffer_get_selection_bounds (GTK_TEXT_BUFFER (buffer), &begin, &end);
 
   if (!gtk_source_search_context_replace (self->context, &begin, &end, replace, -1, &error))
-    g_warning ("Failed to replace match: %s", error->message);
-  else
-    gtk_text_buffer_select_range (GTK_TEXT_BUFFER (buffer), &begin, &begin);
+    {
+      g_warning ("Failed to replace match: %s", error->message);
+      return;
+    }
+
+  gtk_text_buffer_select_range (GTK_TEXT_BUFFER (buffer), &end, &end);
+  _editor_search_bar_move_next (self, FALSE);
 }
 
 void
