@@ -28,6 +28,8 @@
 #include "editor-document.h"
 #include "editor-language-dialog.h"
 #include "editor-page-private.h"
+#include "editor-preferences-dialog-private.h"
+#include "editor-properties-dialog-private.h"
 #include "editor-save-changes-dialog-private.h"
 #include "editor-session-private.h"
 #include "editor-window-private.h"
@@ -153,7 +155,24 @@ editor_window_actions_save_as_cb (GtkWidget  *widget,
   page = editor_window_get_visible_page (self);
 
   if (page != NULL)
-    _editor_page_save_as (page);
+    _editor_page_save_as (page, NULL);
+}
+
+static void
+editor_window_actions_notify_language_cb (EditorDocument       *document,
+                                          GParamSpec           *pspec,
+                                          EditorLanguageDialog *dialog)
+{
+  GtkSourceLanguage *language;
+  const char *language_id = NULL;
+
+  g_assert (EDITOR_IS_DOCUMENT (document));
+  g_assert (EDITOR_IS_LANGUAGE_DIALOG (dialog));
+
+  if ((language = editor_language_dialog_get_language (dialog)))
+    language_id = gtk_source_language_get_id (language);
+
+  _editor_document_persist_syntax_language (document, language_id);
 }
 
 static void
@@ -181,6 +200,17 @@ editor_window_actions_change_language_cb (GtkWidget  *widget,
   g_object_bind_property (document, "language", dialog, "language",
                           G_BINDING_SYNC_CREATE | G_BINDING_BIDIRECTIONAL);
 
+  /* We only want to persist the language to underlying storage if the
+   * user changes the value using the language dialog. Otherwise, we can't
+   * update to better content-type discovery in updates of shared-mime-info
+   * or GtkSourceView as they'd be pinned to the first result we ever get.
+   */
+  g_signal_connect_object (dialog,
+                           "notify::language",
+                           G_CALLBACK (editor_window_actions_notify_language_cb),
+                           document,
+                           G_CONNECT_SWAPPED);
+
   gtk_window_present (GTK_WINDOW (dialog));
 }
 
@@ -358,6 +388,17 @@ editor_window_actions_open_cb (GtkWidget  *widget,
       if (dir != NULL)
         gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (native), dir, NULL);
     }
+  else
+    {
+      g_autoptr(GSettings) settings = g_settings_new ("org.gnome.TextEditor");
+      g_autofree char *uri = g_settings_get_string (settings, "last-save-directory");
+
+      if (uri && uri[0])
+        {
+          g_autoptr(GFile) dir = g_file_new_for_uri (uri);
+          gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (native), dir, NULL);
+        }
+    }
 
   all_files = gtk_file_filter_new ();
   gtk_file_filter_set_name (all_files, _("All Files"));
@@ -520,22 +561,83 @@ editor_window_actions_show_preferences_cb (GtkWidget  *widget,
                                            GVariant   *param)
 {
   EditorWindow *self = (EditorWindow *)widget;
+  GtkWidget *dialog;
 
   g_assert (EDITOR_IS_WINDOW (self));
 
-  gtk_revealer_set_reveal_child (self->preferences_revealer, TRUE);
+  dialog = editor_preferences_dialog_new (EDITOR_APPLICATION_DEFAULT, self);
+  gtk_window_present (GTK_WINDOW (dialog));
 }
 
 static void
-editor_window_actions_hide_preferences_cb (GtkWidget  *widget,
-                                           const char *action_name,
-                                           GVariant   *param)
+editor_window_actions_close_other_pages_cb (GtkWidget  *widget,
+                                            const char *action_name,
+                                            GVariant   *param)
+{
+  EditorWindow *self = (EditorWindow *)widget;
+  EditorPage *current_page;
+  GList *pages;
+
+  g_assert (EDITOR_IS_WINDOW (self));
+
+  current_page = editor_window_get_visible_page (self);
+  pages = _editor_window_get_pages (self);
+  pages = g_list_remove (pages, current_page);
+  _editor_window_request_close_pages (self, pages, TRUE);
+  g_list_free (pages);
+}
+
+static void
+editor_window_actions_page_zoom_in_cb (GtkWidget  *widget,
+                                       const char *action_name,
+                                       GVariant   *param)
 {
   EditorWindow *self = (EditorWindow *)widget;
 
   g_assert (EDITOR_IS_WINDOW (self));
 
-  gtk_revealer_set_reveal_child (self->preferences_revealer, FALSE);
+  _editor_page_zoom_in (editor_window_get_visible_page (self));
+}
+
+static void
+editor_window_actions_page_zoom_out_cb (GtkWidget  *widget,
+                                        const char *action_name,
+                                        GVariant   *param)
+{
+  EditorWindow *self = (EditorWindow *)widget;
+
+  g_assert (EDITOR_IS_WINDOW (self));
+
+  _editor_page_zoom_out (editor_window_get_visible_page (self));
+}
+
+static void
+editor_window_actions_page_zoom_one_cb (GtkWidget  *widget,
+                                        const char *action_name,
+                                        GVariant   *param)
+{
+  EditorWindow *self = (EditorWindow *)widget;
+
+  g_assert (EDITOR_IS_WINDOW (self));
+
+  _editor_page_zoom_one (editor_window_get_visible_page (self));
+}
+
+static void
+editor_window_actions_page_properties_cb (GtkWidget  *widget,
+                                          const char *action_name,
+                                          GVariant   *param)
+{
+  EditorWindow *self = (EditorWindow *)widget;
+  EditorDocument *document;
+  GtkWidget *window;
+
+  g_assert (EDITOR_IS_WINDOW (self));
+
+  document = editor_page_get_document (self->visible_page);
+  window = editor_properties_dialog_new (self, document);
+
+  gtk_window_present (GTK_WINDOW (window));
 }
 
 void
@@ -555,6 +657,10 @@ _editor_window_class_actions_init (EditorWindowClass *klass)
                                    "win.close-page-or-window",
                                    NULL,
                                    editor_window_actions_close_page_cb);
+  gtk_widget_class_install_action (widget_class,
+                                   "win.close-other-pages",
+                                   NULL,
+                                   editor_window_actions_close_other_pages_cb);
   gtk_widget_class_install_action (widget_class,
                                    "win.open",
                                    NULL,
@@ -627,14 +733,26 @@ _editor_window_class_actions_init (EditorWindowClass *klass)
                                    "page.begin-replace",
                                    NULL,
                                    editor_window_actions_begin_replace_cb);
-  gtk_widget_class_install_action (widget_class,
-                                   "win.hide-preferences",
-                                   NULL,
-                                   editor_window_actions_hide_preferences_cb);
   gtk_widget_class_install_action (widget_class,
                                    "win.show-preferences",
                                    NULL,
                                    editor_window_actions_show_preferences_cb);
+  gtk_widget_class_install_action (widget_class,
+                                   "page.zoom-in",
+                                   NULL,
+                                   editor_window_actions_page_zoom_in_cb);
+  gtk_widget_class_install_action (widget_class,
+                                   "page.zoom-out",
+                                   NULL,
+                                   editor_window_actions_page_zoom_out_cb);
+  gtk_widget_class_install_action (widget_class,
+                                   "page.zoom-one",
+                                   NULL,
+                                   editor_window_actions_page_zoom_one_cb);
+  gtk_widget_class_install_action (widget_class,
+                                   "page.properties",
+                                   NULL,
+                                   editor_window_actions_page_properties_cb);
 }
 
 void
@@ -643,12 +761,13 @@ _editor_window_actions_init (EditorWindow *self)
   g_autoptr(GSimpleActionGroup) group = NULL;
   g_autoptr(GSettings) settings = NULL;
   static const gchar *setting_keys[] = {
-    "style-variant",
+    "auto-indent",
     "discover-settings",
     "indent-style",
     "show-line-numbers",
     "show-right-margin",
     "spellcheck",
+    "style-variant",
     "tab-width",
     "wrap-text",
   };
@@ -697,13 +816,18 @@ _editor_window_actions_update (EditorWindow *self,
     }
 
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "win.close-current-page", has_page);
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "win.close-other-pages", has_page);
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.change-language", has_page);
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.discard-changes", externally_modified || (modified && !draft));
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.print", has_page);
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.properties", has_page);
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.save", can_save);
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.save-as", has_page);
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.copy-all", has_page);
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.begin-replace", has_page);
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.begin-search", has_page);
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.zoom-in", has_page);
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.zoom-out", has_page);
+  gtk_widget_action_set_enabled (GTK_WIDGET (self), "page.zoom-one", has_page);
   gtk_widget_action_set_enabled (GTK_WIDGET (self), "win.focus-neighbor", has_page);
 }
